import 'dart:math' show min, pow, max;
import 'dart:ui' show Color;
import 'package:flutter/foundation.dart' show objectRuntimeType;

class LABColor {
  /// Creates a color.
  const LABColor.fromALAB(this.alpha, this.l, this.a, this.b)
      : assert(alpha >= 0.0),
        assert(alpha <= 1.0),
        assert(l >= 0.0),
        assert(l <= 100.0),
        assert(a >= -128.0),
        assert(a <= 127.0),
        assert(b >= -128.0),
        assert(b <= 127.0);

  /// Creates an [LABColor] from an RGB [Color].
  ///
  /// This constructor does not necessarily round-trip with [toColor] because
  /// of floating point imprecision.
  factory LABColor.fromColor(Color color) {
    final double red = color.red / 0xFF;
    final double green = color.green / 0xFF;
    final double blue = color.blue / 0xFF;
    final double alpha = color.alpha / 0xFF;

    final double xLab = 0.412453 * red + 0.357580 * green + 0.180423 * blue;
    final double yLab = 0.212671 * red + 0.715160 * green + 0.072169 * blue;
    final double zLab = 0.019334 * red + 0.119193 * green + 0.950227 * blue;

    const double xN = 95.047;
    const double yN = 100.000;
    const double zN = 108.883;

    double varX = xLab / xN;
    double varY = yLab / yN;
    double varZ = zLab / zN;

    if (varX > 0.008856) {
      varX = pow(varX, 1 / 3).toDouble();
    } else {
      varX = (7.787 * varX) + (16 / 116);
    }

    if (varY > 0.008856) {
      varY = pow(varY, 1 / 3).toDouble();
    } else {
      varY = (7.787 * varY) + (16 / 116);
    }

    if (varZ > 0.008856) {
      varZ = pow(varZ, 1 / 3).toDouble();
    } else {
      varZ = (7.787 * varZ) + (16 / 116);
    }

    final double lLab = (116 * varY) - 16;
    final double aLab = 500 * (varX - varY);
    final double bLab = 200 * (varY - varZ);
    return LABColor.fromALAB(alpha, lLab, aLab, bLab);
  }

  /// Alpha, from 0.0 to 1.0. The describes the transparency of the color.
  /// A value of 0.0 is fully transparent, and 1.0 is fully opaque.
  final double alpha;

  /// Lightness (Lightness), the value range is 0 to 100.
  final double l;

  /// The red and green components (a-axis), range from -128 to 127.
  final double a;

  /// Blue-yellow component (b-axis), with values ranging from -128 to 127.
  final double b;

  /// Returns a copy of this color with the alpha parameter replaced with the
  /// given value.
  LABColor withAlpha(double alpha) => LABColor.fromALAB(alpha, l, a, b);

  /// Returns a copy of this color with the [l] parameter replaced with the
  /// given value.
  LABColor withL(double l) => LABColor.fromALAB(alpha, l, a, b);

  /// Returns a copy of this color with the [a] parameter replaced with
  /// the given value.
  LABColor withA(double a) => LABColor.fromALAB(alpha, l, a, b);

  /// Returns a copy of this color with the [b] parameter replaced with
  /// the given value.
  LABColor withB(double b) => LABColor.fromALAB(alpha, l, a, b);

  /// Returns this color in RGB.
  Color toColor() {
    final double y = (l + 16) / 116;
    final double x = a / 500 + y;
    final double z = y - b / 200;
    final double rRGB = 3.2406 * x - 1.5372 * y - 0.4986 * z;
    final double gRGB = -0.9689 * x + 1.8758 * y + 0.0415 * z;
    final double bRGB = 0.0557 * x - 0.2040 * y + 1.0570 * z;
    return Color.fromARGB(
        (alpha * 0xFF).round(),
        max(0, min(255, (rRGB * 0xFF).round())),
        max(0, min(255, (gRGB * 0xFF).round())),
        max(0, min(255, (bRGB * 0xFF).round())));
  }

  @override
  bool operator ==(Object other) {
    if (identical(this, other)) {
      return true;
    }
    return other is LABColor &&
        other.alpha == alpha &&
        other.l == l &&
        other.a == a &&
        other.b == b;
  }

  @override
  int get hashCode => Object.hash(alpha, l, a, b);

  @override
  String toString() =>
      '${objectRuntimeType(this, 'HSVColor')}($alpha, $l, $a, $b)';
}
