import 'dart:math';
import 'package:idkit/idkit.dart';
import 'package:flutter/material.dart' show Color, HSLColor, HSVColor;

extension IDKitColor on Color {
  /// Generate HSL type colors from RGB type colors.
  HSLColor get hslColor => HSLColor.fromColor(this);

  /// Generate HSV type colors from RGB type colors.
  HSVColor get hsvColor => HSVColor.fromColor(this);

  /// Generate LAB type colors from RGB type colors.
  LABColor get labColor => LABColor.fromColor(this);

  /// Calculate the difference of colors in RGB format using Euclidean distance.
  ///
  /// @color:
  /// [color] is the color in RGB format.
  ///
  /// @tolerance:
  /// [tolerance] is the threshold for judging whether there is a difference in color
  bool euclideanDistanceColorDifference(
    Color color, {
    double tolerance = 1.0,
  }) {
    final int deltaR = red - color.red;
    final int deltaG = green - color.green;
    final int deltaB = blue - color.blue;
    final double averageR = (red + color.red) * 0.5;
    final double deltaC = sqrt((2 + averageR / 256) * pow(deltaR, 2) +
        4 * pow(deltaG, 2) +
        (2 + (255 - averageR) / 256) * pow(deltaB, 2));
    final double deltaE =
        deltaC / sqrt(pow(255, 2) + pow(255, 2) + pow(255, 2));
    return deltaE > tolerance;
  }

  /// Get the UInt32 value of RGBA type color.
  int get rgbaUInt32 {
    final int r = red & 0x000000ff;
    final int g = (green << 8) & 0x0000ff00;
    final int b = (blue << 16) & 0x00ff0000;
    final int a = (alpha << 24) & 0xff000000;
    return a + r + b + g;
  }
}

/// LAB type color extension.
extension IDKitLabColor on LABColor {
  /// Calculate color difference using cie76 standard.
  ///
  /// @color:
  /// [color] is a color of type LAB.
  ///
  /// @tolerance:
  /// [tolerance] is the tolerance value between two colors, the default is 2.3
  /// which is the smallest perceivable difference for human vision.
  bool cie76ColorDifference(
    LABColor color, {
    double tolerance = 2.3,
  }) {
    final double deltaL = l - color.l;
    final double deltaA = a - color.a;
    final double deltaB = b - color.b;
    final double deltaE =
        sqrt(pow(deltaL, 2) + pow(deltaA, 2) + pow(deltaB, 2));
    return deltaE > tolerance;
  }

  /// Compute color differences using the cie94 algorithm.
  ///
  /// @color:
  /// [color] is a color of type LAB.
  ///
  /// @tolerance:
  /// [tolerance] is the threshold for judging whether there is a difference.
  ///
  /// @colorImageStyle:
  /// [colorImageStyle] is an enumeration type of the color image style.
  /// For more detailed content, see [ColorImageStyle].
  ///
  /// @kc:
  /// [kc] is a constant that adjusts the weight for color difference perception.
  /// The value of c depends on the color system used. When using the Lab* color space, the value of kc is usually 1.
  /// If other color spaces are used, it may be necessary to use different values of kc to obtain a more accurate assessment of color differences.
  ///
  /// @kh:
  /// [kh] is a constant that adjusts the weight for brightness difference perception.
  /// [kh] is mainly used to adjust the effect of brightness changes on the evaluation of color differences. It measures the relationship between changes in brightness and how sensitive the human eye is to changes in color.
  /// The value of kh depends on the color system used. When using the Lab* color space, the value of [kh] is usually 1. If you use other color spaces, you may need to use different kh values to get a more accurate assessment of color differences.
  ///
  /// @sl:
  /// [sl] is a constant that adjusts the influence of the L* component (luminance).
  /// The value of sl is usually 1, which is mainly used to adjust the influence of the L* component to better match the perception of the human visual system. Since the perception of brightness by human eyes is more sensitive than that of chroma, it is necessary to accurately evaluate the perception of color difference by adjusting the value of sl.
  bool cie94ColorDifference(
    LABColor color, {
    double tolerance = 1.0,
    ColorImageStyle colorImageStyle = ColorImageStyle.graphicArt,
    double kc = 1.0,
    double kh = 1.0,
    double sl = 1.0,
  }) {
    final double deltaL = l - color.l;
    final double c1 = sqrt(pow(a, 2) + pow(b, 2));
    final double c2 = sqrt(pow(color.a, 2) + pow(color.b, 2));
    final double deltaC12 = c1 - c2;
    final double deltaA = a - color.a;
    final double deltaB = b - color.b;
    final double deltaH =
        sqrt(pow(deltaA, 2) + pow(deltaB, 2) - pow(deltaC12, 2));
    final double sc = 1 + colorImageStyle.k1 * c1;
    final double sh = 1 + colorImageStyle.k2 * c1;
    final double deltaE = sqrt(pow(deltaL / (colorImageStyle.kl * sl), 2) +
        pow(deltaC12 / (kc * sc), 2) +
        pow(deltaH / (kh * sh), 2));
    return deltaE > tolerance;
  }

  /// Compute color differences using the ciede2000 algorithm.
  ///
  /// @color:
  /// [color] is a color of type LAB.
  ///
  /// @tolerance:
  /// [tolerance] is the threshold for judging whether there is a difference.
  ///
  /// @kc:
  /// [kc] refers to the influence coefficient of the degree of chromaticity change on chromatic aberration.
  /// The value range of kC is between 1.0 and 2.0. The role of kC is to reduce the influence on chromatic aberration in the low-saturation area (both C1 and C2 are relatively small), and increase the influence on chromatic aberration in the high-saturation area (both C1 and C2 are relatively large).
  ///
  /// @kh:
  /// [kh] refers to the influence coefficient of the sampling interval of the light source on the color difference, which is used to adjust the influence of the light source change in the color difference.
  /// The function of kh is to adjust the error generated when measuring color difference under different light source conditions, so that the color difference measured under different light sources is more consistent. The value range of kh is between 1.0 and 1.3.
  ///
  /// @kl:
  /// [kl] is the tuning parameter of the photometric reflectance weighting function.
  /// kL is mainly used to control the influence of brightness on chromatic aberration, and the value range of kL is between 1.0 and 2.0.
  bool ciede2000ColorDifference(
    LABColor color, {
    double tolerance = 2.3,
    double kc = 1.0,
    double kh = 1.0,
    double kl = 1.0,
  }) {
    final double deltaL = l - color.l;
    final double averageL = (l + color.l) * 0.5;
    final double c1X = sqrt(pow(a, 2) + pow(b, 2));
    final double c2X = sqrt(pow(color.a, 2) + pow(color.b, 2));
    final double averageCx = (c1X + c2X) * 0.5;
    final double factor =
        sqrt(pow(averageCx, 7) / (pow(averageCx, 7) + pow(25, 7)));
    final double a1P = a + a * 0.5 * (1 - factor);
    final double a2P = color.a + color.a * 0.5 * (1 - factor);
    final double c1P = sqrt(pow(a1P, 2) + pow(b, 2));
    final double c2P = sqrt(pow(a2P, 2) + pow(color.b, 2));
    final double averageCP = (c1P + c2P) * 0.5;
    final double deltaCP = c1P - c2P;
    double deltaHP = 0;
    double averageHP = 0;
    if (c1P != 0 && c2P != 0) {
      final double h1P = atan2(b, a1P).range(min: 0, max: 360);
      final double h2P = atan2(color.b, a2P).range(min: 0, max: 360);
      final double deltaH = (h1P - h2P).abs();

      /// deltaHP
      if (deltaH <= 180) {
        deltaHP = h2P - h1P;
      }
      if (deltaH > 180 && h2P <= h1P) {
        deltaHP = h2P - h1P + 360;
      }
      if (deltaH > 180 && h2P > h1P) {
        deltaHP = h2P - h1P - 360;
      }

      /// averageH
      if (deltaH > 180) {
        averageHP = (h1P + h2P + 360) * 0.5;
      }
      if (deltaH <= 180) {
        averageHP = (h1P + h2P) * 0.5;
      }
    }
    final double deltaDHP = 2 * sqrt(c1P * c2P) * sin(deltaHP * 0.5);
    final double t = 1 -
        0.17 * cos(averageHP - 30) +
        0.24 * cos(2 * averageHP) +
        0.32 * cos(3 * averageHP + 6) -
        0.2 * cos(4 * averageHP - 63);
    final double sl =
        1 + (0.015 * pow(averageL - 50, 2)) / sqrt(20 + pow(averageL - 50, 2));
    final double sc = 1 + 0.045 * averageCP;
    final double sh = 1 + 0.015 * averageCP * t;
    final double rt =
        -2 * factor * sin(60 * exp(-pow((averageHP - 275) / 25, 2)));
    final double deltaE = sqrt(
      pow(deltaL / (kl * sl), 2) +
          pow(deltaCP / (kc * sc), 2) +
          pow(deltaDHP / (kh * sh), 2) +
          rt * (deltaCP / (kc * sc)) * (deltaDHP / (kh * sh)),
    );
    return deltaE > tolerance;
  }
}
