import 'dart:async';
import 'dart:math';
import 'dart:typed_data';
import 'dart:ui';
import 'package:idkit/idkit.dart';

/// Some Convenience Handling of Fultter Images.
extension IImage on Image {
  /// Image size.
  Size get size => Size(width * 1, height * 1);

  /// Image rectangle with coordinates at (0,0) point.
  Rect rect({Offset offset = Offset.zero}) => offset & size;

  /// Get the Uint8List data of the image.
  Future<Uint8List?> uint8List(
      {ImageByteFormat format = ImageByteFormat.rawRgba}) async {
    final ByteData? byteData = await toByteData(format: format);
    return byteData?.buffer.asUint8List();
  }

  /// Get the Uint8List data of the image.
  Future<Uint32List?> uint32List(
      {ImageByteFormat format = ImageByteFormat.rawRgba}) async {
    final ByteData? byteData = await toByteData(format: format);
    return byteData?.buffer.asUint32List();
  }

  /// Change image size.
  Image changeSizeSync({
    int? imgW,
    int? imgH,
    bool isDispose = false,
  }) {
    final int imgW0 = imgW ?? width;
    final int imgH0 = imgH ?? height;
    final Rect rect = Offset.zero & Size(imgW0 * 1, imgH0 * 1);
    final PictureRecorder pictureRecorder = PictureRecorder();
    final Canvas canvas = Canvas(pictureRecorder, rect);
    canvas.drawImageRect(this, Offset.zero & size, rect,
        Paint()..filterQuality = FilterQuality.high);
    final Picture picture = pictureRecorder.endRecording();
    if (isDispose) dispose();
    return picture.toImageSync(imgW0, imgH0);
  }

  /// Resize an image asynchronously.
  Future<Image> changeSize({
    int? imgW,
    int? imgH,
    bool isDispose = false,
  }) {
    final int imgW0 = imgW ?? width;
    final int imgH0 = imgH ?? height;
    final Rect rect = Offset.zero & Size(imgW0 * 1, imgH0 * 1);
    final PictureRecorder pictureRecorder = PictureRecorder();
    final Canvas canvas = Canvas(pictureRecorder, rect);
    canvas.drawImageRect(this, Offset.zero & size, rect,
        Paint()..filterQuality = FilterQuality.high);
    final Picture picture = pictureRecorder.endRecording();
    if (isDispose) dispose();
    return picture.toImage(imgW0, imgH0);
  }

  /// Image scaling.
  Image scaleSync({
    double scale = 1,
    bool isDispose = false,
  }) {
    final int imgW0 = (width * scale).toInt();
    final int imgH0 = (height * scale).toInt();
    final Rect rect = Offset.zero & Size(imgW0 * 1, imgH0 * 1);
    final PictureRecorder pictureRecorder = PictureRecorder();
    final Canvas canvas = Canvas(pictureRecorder, rect);
    canvas.drawImageRect(this, Offset.zero & size, rect,
        Paint()..filterQuality = FilterQuality.high);
    final Picture picture = pictureRecorder.endRecording();
    if (isDispose) dispose();
    return picture.toImageSync(imgW0, imgH0);
  }

  /// Image scaling.
  Future<Image> scale({
    double scale = 1,
    bool isDispose = false,
  }) {
    final int imgW0 = (width * scale).toInt();
    final int imgH0 = (height * scale).toInt();
    final Rect rect = Offset.zero & Size(imgW0 * 1, imgH0 * 1);
    final PictureRecorder pictureRecorder = PictureRecorder();
    final Canvas canvas = Canvas(pictureRecorder, rect);
    canvas.drawImageRect(this, Offset.zero & size, rect,
        Paint()..filterQuality = FilterQuality.high);
    final Picture picture = pictureRecorder.endRecording();
    if (isDispose) dispose();
    return picture.toImage(imgW0, imgH0);
  }

  /// Image decolorization.
  Future<Image?> removalColor({
    Color color = const Color(0xffffffff),
    double tolerance = 1,
  }) async {
    Completer<Image?> completer = Completer();
    final ByteData? byteData =
        await toByteData(format: ImageByteFormat.rawRgba);
    if (byteData.isNoNull) {
      final Uint32List uint32list = byteData!.buffer.asUint32List();
      uint32list.iFor((index, element) {
        final bool isDiff = element.uInt32toColor
            .euclideanDistanceColorDifference(color, tolerance: tolerance);
        if (!isDiff) uint32list[index] = 0;
      });
      decodeImageFromPixels(
        uint32list.buffer.asUint8List(),
        width,
        height,
        PixelFormat.rgba8888,
        (Image result) => completer.complete(result),
      );
    }
    return completer.future;
  }

  /// Get the real content size of the image.
  Future<Rect> realContentSize({
    Color ignoreColor = const Color(0xffffffff),
    double tolerance = 1,
  }) async {
    Completer<Rect> completer = Completer();
    final ByteData? byteData =
        await toByteData(format: ImageByteFormat.rawRgba);
    if (byteData.isNull) {
      completer.complete(rect());
    } else {
      final Uint32List uint32list = byteData!.buffer.asUint32List();
      late double minDx = -1;
      late double minDy = -1;
      late double maxDx = -1;
      late double maxDy = -1;
      bool isAssign = false;
      uint32list.iFor((index, element) {
        final bool isDiff = element.uInt32toColor
            .euclideanDistanceColorDifference(ignoreColor,
                tolerance: tolerance);
        if (isDiff) {
          final int dx = index % width;
          final int dy = index ~/ width;
          if (!isAssign) {
            isAssign = true;
            minDx = dx * 1;
            maxDx = dx * 1;
            minDy = dy * 1;
            maxDy = dy * 1;
          } else {
            minDx = min(dx * 1, minDx);
            minDy = min(dy * 1, minDy);
            maxDx = max(dx * 1, maxDx);
            maxDy = max(dy * 1, maxDy);
          }
        }
      });
      final Offset start = Offset(minDx, minDy);
      final Size realSize = Size((maxDx - minDx), (maxDy - minDy));
      completer.complete(start & realSize);
    }
    return completer.future;
  }

  /// Get the real content in the image.
  Future<Image> realContentImage({
    Color ignoreColor = const Color(0xffffffff),
    bool includeIgnoreColor = false,
    double tolerance = 1,
    bool isDispose = false,
  }) async {
    Completer<Image> completer = Completer();
    final ByteData? byteData =
        await toByteData(format: ImageByteFormat.rawRgba);
    if (byteData.isNull) {
      completer.complete(this);
    } else {
      final Uint32List uint32list = byteData!.buffer.asUint32List();
      late double minDx = -1;
      late double minDy = -1;
      late double maxDx = -1;
      late double maxDy = -1;
      bool isAssign = false;
      uint32list.iFor((index, element) {
        final bool isDiff = element.uInt32toColor
            .euclideanDistanceColorDifference(ignoreColor,
                tolerance: tolerance);
        if (isDiff) {
          final int dx = index % width;
          final int dy = index ~/ width;
          if (!isAssign) {
            isAssign = true;
            minDx = dx * 1;
            maxDx = dx * 1;
            minDy = dy * 1;
            maxDy = dy * 1;
          } else {
            minDx = min(dx * 1, minDx);
            minDy = min(dy * 1, minDy);
            maxDx = max(dx * 1, maxDx);
            maxDy = max(dy * 1, maxDy);
          }
        } else if (!includeIgnoreColor) {
          uint32list[index] = 0;
        }
      });
      final Offset start = Offset(minDx, minDy);
      final Size realSize = Size((maxDx - minDx), (maxDy - minDy));
      if (includeIgnoreColor) {
        final PictureRecorder recorder = PictureRecorder();
        final Canvas canvas = Canvas(recorder);
        canvas.drawImageRect(this, start & realSize, Offset.zero & realSize,
            Paint()..filterQuality = FilterQuality.high);
        completer.complete(recorder
            .endRecording()
            .toImageSync(realSize.width.toInt(), realSize.height.toInt()));
      } else {
        decodeImageFromPixels(
          uint32list.buffer.asUint8List(),
          width,
          height,
          PixelFormat.rgba8888,
          (Image result) {
            final PictureRecorder recorder = PictureRecorder();
            final Canvas canvas = Canvas(recorder);
            canvas.drawImageRect(
                result,
                start & realSize,
                Offset.zero & realSize,
                Paint()..filterQuality = FilterQuality.high);
            completer.complete(recorder
                .endRecording()
                .toImageSync(realSize.width.toInt(), realSize.height.toInt()));
          },
        );
      }
    }
    if (isDispose) dispose();
    return completer.future;
  }
}
