import 'dart:async';
import 'dart:collection';
import 'dart:typed_data' show Uint32List;
import 'dart:ui' show Image, PixelFormat, decodeImageFromPixels;
import 'package:flutter/material.dart' show Offset, Color;
import 'package:idkit/src/class/enums.dart';
import 'package:idkit/src/class/int_offset.dart';
import 'package:idkit/src/class/lab_color.dart';
import 'package:idkit/src/function/color/func_color.dart';
import 'package:idkit/src/function/image/func_image.dart';
import 'package:idkit/src/function/int/func_int.dart';

/// Flooding of images.
extension FloodingImage on Image {
  /// Image Euclidean distance flooding.
  ///
  /// @[offset]
  /// The starting point for image flooding.
  ///
  /// @[fillColor]
  /// The fill color of the image flood, in RGB format.
  ///
  /// @[tolerance]
  /// Color tolerance for image flooding, defaults to 1.
  ///
  /// @[isDispose]
  /// Whether to release the original image after image flooding.
  Future<Image> euclideanDistanceFlooding({
    required Offset offset,
    required Color fillColor,
    double tolerance = 1.0,
    bool isDispose = false,
  }) async {
    final Uint32List? uint32list = await uint32List();
    if (uint32list == null || uint32list.isEmpty) return this;
    final int dx = offset.dx.floor();
    final int dy = offset.dy.floor();
    if (dx < 0 || dx == width || dy < 0 || dy == height) return this;
    final int index = (width * dy + dx).clamp(0, uint32list.length - 1);
    final Color collectionColor = uint32list[index].uInt32toColor;
    if (!collectionColor.euclideanDistanceColorDifference(fillColor,
        tolerance: tolerance)) return this;
    final Completer<Image> completer = Completer();
    _floodingMethod(
        uint32list,
        fillColor.rgbaUInt32,
        width,
        height,
        IntOffset(dx, dy),
        (color) => collectionColor.euclideanDistanceColorDifference(color,
            tolerance: tolerance));
    decodeImageFromPixels(
        uint32list.buffer.asUint8List(), width, height, PixelFormat.rgba8888,
        (result) {
      completer.complete(result);
    });
    if (isDispose) dispose();
    return completer.future;
  }

  /// Image cie76 flood.
  ///
  /// @[offset]
  /// The starting point for image flooding.
  ///
  /// @[fillColor]
  /// The fill color of the image flood, in RGB format.
  ///
  /// @[tolerance]
  /// Color tolerance for image flooding, defaults to 2.3.
  ///
  /// @[isDispose]
  /// Whether to release the original image after image flooding.
  Future<Image> cie76Flooding({
    required Offset offset,
    required Color fillColor,
    double tolerance = 2.3,
    bool isDispose = false,
  }) async {
    final Uint32List? uint32list = await uint32List();
    if (uint32list == null || uint32list.isEmpty) return this;
    final int dx = offset.dx.floor();
    final int dy = offset.dy.floor();
    if (dx < 0 || dx == width || dy < 0 || dy == height) return this;
    final int index = (width * dy + dx).clamp(0, uint32list.length - 1);
    final LABColor collectionColor = uint32list[index].uInt32toColor.labColor;
    if (!collectionColor.cie76ColorDifference(fillColor.labColor,
        tolerance: tolerance)) return this;
    final Completer<Image> completer = Completer();
    _floodingMethod(
        uint32list,
        fillColor.rgbaUInt32,
        width,
        height,
        IntOffset(dx, dy),
        (color) => collectionColor.cie76ColorDifference(color.labColor,
            tolerance: tolerance));
    decodeImageFromPixels(
        uint32list.buffer.asUint8List(), width, height, PixelFormat.rgba8888,
        (result) {
      completer.complete(result);
    });
    if (isDispose) dispose();
    return completer.future;
  }

  /// Image cie94 flood.
  ///
  /// @[offset]
  /// The starting point for image flooding.
  ///
  /// @[fillColor]
  /// The fill color of the image flood, in RGB format.
  ///
  /// @colorImageStyle:
  /// [colorImageStyle] is an enumeration type of the color image style.
  /// For more detailed content, see [ColorImageStyle].
  ///
  /// @[tolerance]
  /// Color tolerance for image flooding, defaults to 2.3.
  ///
  /// @kc:
  /// [kc] is a constant that adjusts the weight for color difference perception.
  /// The value of c depends on the color system used. When using the Lab* color space, the value of kc is usually 1.
  /// If other color spaces are used, it may be necessary to use different values of kc to obtain a more accurate assessment of color differences.
  ///
  /// @kh:
  /// [kh] is a constant that adjusts the weight for brightness difference perception.
  /// [kh] is mainly used to adjust the effect of brightness changes on the evaluation of color differences. It measures the relationship between changes in brightness and how sensitive the human eye is to changes in color.
  /// The value of kh depends on the color system used. When using the Lab* color space, the value of [kh] is usually 1. If you use other color spaces, you may need to use different kh values to get a more accurate assessment of color differences.
  ///
  /// @sl:
  /// [sl] is a constant that adjusts the influence of the L* component (luminance).
  /// The value of sl is usually 1, which is mainly used to adjust the influence of the L* component to better match the perception of the human visual system. Since the perception of brightness by human eyes is more sensitive than that of chroma, it is necessary to accurately evaluate the perception of color difference by adjusting the value of sl.
  ///
  /// @[isDispose]
  /// Whether to release the original image after image flooding.
  Future<Image> cie94Flooding({
    required Offset offset,
    required Color fillColor,
    ColorImageStyle colorImageStyle = ColorImageStyle.graphicArt,
    double tolerance = 2.3,
    double kc = 1.0,
    double kh = 1.0,
    double sl = 1.0,
    bool isDispose = false,
  }) async {
    final Uint32List? uint32list = await uint32List();
    if (uint32list == null || uint32list.isEmpty) return this;
    final int dx = offset.dx.floor();
    final int dy = offset.dy.floor();
    if (dx < 0 || dx == width || dy < 0 || dy == height) return this;
    final int index = (width * dy + dx).clamp(0, uint32list.length - 1);
    final LABColor collectionColor = uint32list[index].uInt32toColor.labColor;
    if (!collectionColor.cie94ColorDifference(
      fillColor.labColor,
      colorImageStyle: colorImageStyle,
      tolerance: tolerance,
      kc: kc,
      kh: kh,
      sl: sl,
    )) return this;
    final Completer<Image> completer = Completer();
    _floodingMethod(
      uint32list,
      fillColor.rgbaUInt32,
      width,
      height,
      IntOffset(dx, dy),
      (color) => collectionColor.cie94ColorDifference(
        color.labColor,
        colorImageStyle: colorImageStyle,
        tolerance: tolerance,
        kc: kc,
        kh: kh,
        sl: sl,
      ),
    );
    decodeImageFromPixels(
        uint32list.buffer.asUint8List(), width, height, PixelFormat.rgba8888,
        (result) {
      completer.complete(result);
    });
    if (isDispose) dispose();
    return completer.future;
  }

  /// Image ciede2000 flood.
  ///
  /// @[offset]
  /// The starting point for image flooding.
  ///
  /// @[fillColor]
  /// The fill color of the image flood, in RGB format.
  ///
  /// @[tolerance]
  /// Color tolerance for image flooding, defaults to 2.3.
  ///
  /// @kc:
  /// [kc] is a constant that adjusts the weight for color difference perception.
  /// The value of c depends on the color system used. When using the Lab* color space, the value of kc is usually 1.
  /// If other color spaces are used, it may be necessary to use different values of kc to obtain a more accurate assessment of color differences.
  ///
  /// @kh:
  /// [kh] is a constant that adjusts the weight for brightness difference perception.
  /// [kh] is mainly used to adjust the effect of brightness changes on the evaluation of color differences. It measures the relationship between changes in brightness and how sensitive the human eye is to changes in color.
  /// The value of kh depends on the color system used. When using the Lab* color space, the value of [kh] is usually 1. If you use other color spaces, you may need to use different kh values to get a more accurate assessment of color differences.
  ///
  /// @kl:
  /// [kl] is the tuning parameter of the photometric reflectance weighting function.
  /// kL is mainly used to control the influence of brightness on chromatic aberration, and the value range of kL is between 1.0 and 2.0.
  ///
  /// @[isDispose]
  /// Whether to release the original image after image flooding.
  Future<Image> ciede2000Flooding({
    required Offset offset,
    required Color fillColor,
    double tolerance = 2.3,
    double kc = 1.0,
    double kh = 1.0,
    double kl = 1.0,
    bool isDispose = false,
  }) async {
    final Uint32List? uint32list = await uint32List();
    if (uint32list == null || uint32list.isEmpty) return this;
    final int dx = offset.dx.floor();
    final int dy = offset.dy.floor();
    if (dx < 0 || dx == width || dy < 0 || dy == height) return this;
    final int index = (width * dy + dx).clamp(0, uint32list.length - 1);
    final LABColor collectionColor = uint32list[index].uInt32toColor.labColor;
    if (!collectionColor.ciede2000ColorDifference(
      fillColor.labColor,
      tolerance: tolerance,
      kc: kc,
      kh: kh,
      kl: kl,
    )) return this;
    final Completer<Image> completer = Completer();
    _floodingMethod(
      uint32list,
      fillColor.rgbaUInt32,
      width,
      height,
      IntOffset(dx, dy),
      (color) => collectionColor.ciede2000ColorDifference(
        color.labColor,
        tolerance: tolerance,
        kc: kc,
        kh: kh,
        kl: kl,
      ),
    );
    decodeImageFromPixels(
        uint32list.buffer.asUint8List(), width, height, PixelFormat.rgba8888,
        (result) {
      completer.complete(result);
    });
    if (isDispose) dispose();
    return completer.future;
  }

  /// image flooding.
  void _floodingMethod(
    Uint32List uint32list,
    int fcValue,
    int width,
    int height,
    IntOffset cOffset,
    bool Function(Color color) colorDiff,
  ) {
    int fdx = 0;
    bool spanAbove = false;
    bool spanBelow = false;
    final ListQueue<IntOffset> queue = ListQueue(2);
    queue.add(cOffset);
    while (queue.isNotEmpty) {
      final IntOffset offset = queue.removeFirst();
      fdx = offset.dx;
      final fdy = offset.dy;
      while (fdx > 0 &&
          fdy < height &&
          !colorDiff.call(uint32list[fdy * width + fdx].uInt32toColor)) {
        fdx--;
      }
      spanAbove = spanBelow = false;
      fdx++;
      while (fdx < width &&
          fdy < height &&
          !colorDiff.call(uint32list[fdy * width + fdx].uInt32toColor)) {
        uint32list[fdy * width + fdx] = fcValue;
        if (!spanAbove &&
            fdy > 0 &&
            !colorDiff
                .call(uint32list[(fdy - 1) * width + fdx].uInt32toColor)) {
          queue.add(IntOffset(fdx, fdy - 1));
          spanAbove = true;
        } else if (spanAbove &&
            fdy > 0 &&
            colorDiff.call(uint32list[(fdy - 1) * width + fdx].uInt32toColor)) {
          spanAbove = false;
        }
        if (!spanBelow &&
            fdy < height - 1 &&
            !colorDiff
                .call(uint32list[(fdy + 1) * width + fdx].uInt32toColor)) {
          queue.add(IntOffset(fdx, fdy + 1));
          spanBelow = true;
        } else if (spanBelow &&
            fdy < height - 1 &&
            colorDiff.call(uint32list[(fdy + 1) * width + fdx].uInt32toColor)) {
          spanBelow = false;
        }
        fdx++;
      }
    }
  }
}
